from django.core import checks
from django.db import models


class CustomFieldMixin(object):
    def __init__(self, *args, **kwargs):
        self.custom_indexes = kwargs.pop(u'custom_indexes', None)
        super(CustomFieldMixin, self).__init__(*args, **kwargs)

    def deconstruct(self):
        name, path, args, kwargs = super(CustomFieldMixin, self).deconstruct()
        if self.custom_indexes:
            kwargs[u'custom_indexes'] = self.custom_indexes
        return name, path, args, kwargs

    def check(self, **kwargs):
        errors = super(CustomFieldMixin, self).check(**kwargs)
        errors.extend(self._check_custom_indexes())
        return errors

    def _check_custom_indexes(self):
        if not self.custom_indexes:
            return []
        if not isinstance(self.custom_indexes, list):
            return [
                checks.Error(
                    "'custom_indexes' must be a list.",
                    hint=None,
                    obj=self,
                    id='django_custom_indexes.E001',
                )
            ]
        for index in self.custom_indexes:
            if not isinstance(index, dict):
                return [
                    checks.Error(
                        "Each index must be a dict.",
                        hint=None,
                        obj=self,
                        id='django_custom_indexes.E002',
                    )
                ]
            if not index.get('columns') and not index.get('name'):
                return [
                    checks.Error(
                        "Each index must either have a name or a list of columns",
                        hint=None,
                        obj=self,
                        id='django_custom_indexes.E003',
                    )
                ]
        return []


class CustomAutoField(CustomFieldMixin, models.AutoField):
    pass


class CustomBigIntegerField(CustomFieldMixin, models.BigIntegerField):
    pass


class CustomBinaryField(CustomFieldMixin, models.BinaryField):
    pass


class CustomBooleanField(CustomFieldMixin, models.BooleanField):
    pass


class CustomCharField(CustomFieldMixin, models.CharField):
    pass


class CustomCommaSeparatedIntegerField(CustomFieldMixin, models.CommaSeparatedIntegerField):
    pass


class CustomDateField(CustomFieldMixin, models.DateField):
    pass


class CustomDateTimeField(CustomFieldMixin, models.DateTimeField):
    pass


class CustomDecimalField(CustomFieldMixin, models.DecimalField):
    pass


class CustomDurationField(CustomFieldMixin, models.DurationField):
    pass


class CustomEmailField(CustomFieldMixin, models.EmailField):
    pass


class CustomFileField(CustomFieldMixin, models.FileField):
    pass


class CustomFilePathField(CustomFieldMixin, models.FilePathField):
    pass


class CustomFloatField(CustomFieldMixin, models.FloatField):
    pass


class CustomForeignKey(CustomFieldMixin, models.ForeignKey):
    pass


class CustomForeignObject(CustomFieldMixin, models.ForeignObject):
    pass


class CustomGenericIPAddressField(CustomFieldMixin, models.GenericIPAddressField):
    pass


class CustomIPAddressField(CustomFieldMixin, models.IPAddressField):
    pass


class CustomImageField(CustomFieldMixin, models.ImageField):
    pass


class CustomIntegerField(CustomFieldMixin, models.IntegerField):
    pass


class CustomManyToManyField(CustomFieldMixin, models.ManyToManyField):
    pass


class CustomNullBooleanField(CustomFieldMixin, models.NullBooleanField):
    pass


class CustomOneToOneField(CustomFieldMixin, models.OneToOneField):
    pass


class CustomOrderWrt(CustomFieldMixin, models.OrderWrt):
    pass


class CustomPositiveIntegerField(CustomFieldMixin, models.PositiveIntegerField):
    pass


class CustomPositiveSmallIntegerField(CustomFieldMixin, models.PositiveSmallIntegerField):
    pass


class CustomSlugField(CustomFieldMixin, models.SlugField):
    pass


class CustomSmallIntegerField(CustomFieldMixin, models.SmallIntegerField):
    pass


class CustomTextField(CustomFieldMixin, models.TextField):
    pass


class CustomTimeField(CustomFieldMixin, models.TimeField):
    pass


class CustomURLField(CustomFieldMixin, models.URLField):
    pass


class CustomUUIDField(CustomFieldMixin, models.UUIDField):
    pass



import importlib

from django.conf import settings

try:
    # Django >= 1.9
    from django.db.backends.postgresql.base import *

    DEFAULT_BACKEND = 'django.db.backends.postgresql'
except ImportError:
    DEFAULT_BACKEND = 'django.db.backends.postgresql_psycopg2'

BASE_ENGINE = getattr(settings, 'DJANGO_CUSTOM_INDEXES_BASE_ENGINE', DEFAULT_BACKEND)

BASE_MODULE = importlib.import_module('{}.base'.format(BASE_ENGINE))


class DatabaseSchemaEditor(BASE_MODULE.DatabaseWrapper.SchemaEditorClass):
    sql_create_custom_index = """
        CREATE %(unique)s INDEX %(name)s ON %(table)s %(using)s %(columns)s %(where)s
    """

    def add_indexes(self, model, indexes):
        for index in indexes:
            columns = ", ".join(index.get('columns', []))
            name = index.get('name')
            if not name:
                name = self.quote_name(self._create_index_name(model, index.get('columns'), suffix="_custom"))
            self.deferred_sql.append(self.sql_create_custom_index % {
                "unique": "UNIQUE" if index.get('unique') else "",
                "name": name,
                "table": self.quote_name(model._meta.db_table),
                "using": index.get('using', ''),
                "columns": "({})".format(columns) if columns else "",
                "where": index.get('where', ''),
            })

    def create_model(self, model):
        super(DatabaseSchemaEditor, self).create_model(model)
        for field in model._meta.local_fields:
            self.add_indexes(model, getattr(field, 'custom_indexes', []))

    def add_field(self, model, field):
        super(DatabaseSchemaEditor, self).add_field(model, field)
        self.add_indexes(model, getattr(field, 'custom_indexes', []))

    def _alter_field(self, model, old_field, new_field, old_type, new_type, old_db_params, new_db_params, strict=False):
        super(DatabaseSchemaEditor, self)._alter_field(
            model, old_field, new_field, old_type, new_type, old_db_params, new_db_params, strict
        )
        old_indexes = getattr(old_field, 'custom_indexes', [])
        new_indexes = getattr(new_field, 'custom_indexes', [])
        create_indexes = []
        delete_indexes = []

        if old_indexes != new_indexes:
            for index in new_indexes:
                if index not in old_indexes:
                    create_indexes.append(index)
            for index in old_indexes:
                if index not in new_indexes:
                    delete_indexes.append(index)
        self.add_indexes(model, create_indexes)

        for index in delete_indexes:
            name = index.get(
                'name', self.quote_name(self._create_index_name(model, index['columns'], suffix="_custom"))
            )
            self.deferred_sql.append(self.sql_delete_index % {"name": name})



import os

from setuptools import setup, find_packages

with open(os.path.join(os.path.dirname(__file__), 'README.rst')) as readme:
    README = readme.read()

# allow setup.py to be run from any path
os.chdir(os.path.normpath(os.path.join(os.path.abspath(__file__), os.pardir)))

setup(
    name='django-custom-indexes',
    url='https://github.com/joshtechnologygroup/django-custom-indexes/',
    packages=find_packages(),
    version='0.1.4',
    description='PostgreSQL custom indexes for Django models',
    long_description=README,
    author='Anshul Chhangani',
    author_email='anshul.chhangani@gmail.com',
    license='BSD',
    install_requires=[],
    classifiers=[
        'Environment :: Web Environment',
        'Framework :: Django :: 1.7',
        'Framework :: Django :: 1.8',
        'Framework :: Django :: 1.9',
        'Framework :: Django :: 1.10',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: BSD License',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 2.7',
        'Topic :: Database',
        'Topic :: Internet :: WWW/HTTP',
    ]
)



import importlib

from django.conf import settings

try:
    # Django >= 1.9
    from django.db.backends.postgresql.base import *

    DEFAULT_BACKEND = 'django.db.backends.postgresql'
except ImportError:
    DEFAULT_BACKEND = 'django.db.backends.postgresql_psycopg2'

from django_custom_indexes.backends.postgresql_psycopg2.schema import DatabaseSchemaEditor


BASE_ENGINE = getattr(settings, 'DJANGO_CUSTOM_INDEXES_BASE_ENGINE', DEFAULT_BACKEND)

BASE_MODULE = importlib.import_module('{}.base'.format(BASE_ENGINE))


class DatabaseWrapper(BASE_MODULE.DatabaseWrapper):
    SchemaEditorClass = DatabaseSchemaEditor
